@@ -6,21 +6,21 @@ inherit from this interface to implement new learners. In addition, a number
 of learners are provided out of the box for testing and baseline comparisons.
 """
 
-from coba.learners.core import Learner, SafeLearner
-from coba.learners.bandit import RandomLearner, EpsilonBanditLearner, UcbBanditLearner
+from coba.learners.core import Learner, FixedLearner, RandomLearner, SafeLearner
+from coba.learners.bandit import EpsilonBanditLearner, UcbBanditLearner
 from coba.learners.corral import CorralLearner
-from coba.learners.vowpal import VowpalLearner
+from coba.learners.vowpal import VowpalLearner, VowpalMediator
 from coba.learners.linUCB import LinUCBLearner
-from coba.learners.regCB import RegCBLearner
 
 __all__ = [
     'Learner',
     'SafeLearner',
     'RandomLearner',
+    'FixedLearner',
     'EpsilonBanditLearner',
     'UcbBanditLearner',
     'CorralLearner',
     'VowpalLearner',
     'LinUCBLearner',
-    'RegCBLearner'
+    'VowpalMediator'
 ]
\ No newline at end of file
