@@ -3,56 +3,12 @@
 import math
 
 from collections import defaultdict
-from typing import Any, Dict, Tuple, Sequence, Optional, cast, Hashable
+from typing import Any, Dict, Sequence, Optional, cast, Hashable
 
-from coba.simulations import Context, Action
+from coba.environments import Context, Action
 from coba.statistics import OnlineVariance
-from coba.learners.core import Learner, Key
+from coba.learners.core import Learner, Probs, Info
 
-class RandomLearner(Learner):
-    """A Learner implementation that selects an action at random and learns nothing."""
-
-    @property
-    def family(self) -> str:
-        """The family of the learner.
-
-        See the base class for more information
-        """  
-        return "random"
-
-    @property
-    def params(self) -> Dict[str, Any]:
-        """The parameters of the learner.
-        
-        See the base class for more information
-        """
-        return {}
-
-    def predict(self, key: Key, context: Context, actions: Sequence[Action]) -> Sequence[float]:
-        """Choose a random action from the action set.
-        
-        Args:
-            key: The key identifying the interaction we are choosing for.
-            context: The context we're currently in. See the base class for more information.
-            actions: The actions to choose from. See the base class for more information.
-
-        Returns:
-            The probability of taking each action. See the base class for more information.
-        """
-        return [1/len(actions)] * len(actions)
-
-    def learn(self, key: Key, context: Context, action: Action, reward: float, probability: float) -> None:
-        """Learns nothing.
-
-        Args:
-            key: The key identifying the interaction this observed reward came from.
-            context: The context we're learning about. See the base class for more information.
-            action: The action that was selected in the context. See the base class for more information.
-            reward: The reward that was gained from the action. See the base class for more information.
-            probability: The probability that the given action was taken.
-        """
-        pass
- 
 class EpsilonBanditLearner(Learner):
     """A lookup table bandit learner with epsilon-greedy exploration."""
 
@@ -69,28 +25,18 @@ class EpsilonBanditLearner(Learner):
         self._N: Dict[Hashable, int            ] = defaultdict(int)
         self._Q: Dict[Hashable, Optional[float]] = defaultdict(int)
 
-    @property
-    def family(self) -> str:
-        """The family of the learner.
-
-        See the base class for more information
-        """
-
-        return "bandit_epsilongreedy"
-
     @property
     def params(self) -> Dict[str, Any]:
         """The parameters of the learner.
         
         See the base class for more information
         """
-        return {"epsilon": self._epsilon }
+        return {"family": "bandit_epsilon", "epsilon": self._epsilon }
 
-    def predict(self, key: Key, context: Context, actions: Sequence[Action]) -> Sequence[float]:
+    def predict(self, context: Context, actions: Sequence[Action]) -> Probs:
         """Determine a PMF with which to select the given actions.
 
         Args:
-            key: The key identifying the interaction we are choosing for.
             context: The context we're currently in. See the base class for more information.
             actions: The actions to choose from. See the base class for more information.
 
@@ -98,37 +44,32 @@ class EpsilonBanditLearner(Learner):
             The probability of taking each action. See the base class for more information.
         """
 
-        keys        = [ self._key(action) for action in actions ]
-        values      = [ self._Q[key] for key in keys ]
+        values      = [ self._Q[action] for action in actions ]
         max_value   = None if set(values) == {None} else max(v for v in values if v is not None)
         max_indexes = [i for i in range(len(values)) if values[i]==max_value]
 
         prob_selected_randomly = [1/len(actions) * self._epsilon] * len(actions)
         prob_selected_greedily = [ int(i in max_indexes)/len(max_indexes) * (1-self._epsilon) for i in range(len(actions))]
 
-        return [p1+p2 for p1,p2 in zip(prob_selected_randomly,prob_selected_greedily)]
+        return [ round(p1+p2,4) for p1,p2 in zip(prob_selected_randomly,prob_selected_greedily)]
 
-    def learn(self, key: Key, context: Context, action: Action, reward: float, probability: float) -> None:
+    def learn(self, context: Context, action: Action, reward: float, probability: float, info: Info) -> None:
         """Learn from the given interaction.
 
         Args:
-            key: The key identifying the interaction this observed reward came from.
             context: The context we're learning about. See the base class for more information.
             action: The action that was selected in the context. See the base class for more information.
             reward: The reward that was gained from the action. See the base class for more information.
-            probability: The probability that the given action was taken.
+            probability: The probability with wich the given action was selected.
+            info: Optional information provided during prediction step for use in learning.
         """
 
-        a_key = self._key(action)
-        alpha = 1/(self._N[a_key]+1)
+        alpha = 1/(self._N[action]+1)
 
-        old_Q = cast(float, 0 if self._Q[a_key] is None else self._Q[a_key])
+        old_Q = cast(float, 0 if self._Q[action] is None else self._Q[action])
 
-        self._Q[a_key] = (1-alpha) * old_Q + alpha * reward
-        self._N[a_key] = self._N[a_key] + 1
-
-    def _key(self, action: Action) -> Hashable:
-        return tuple(action.items()) if isinstance(action,dict) else action
+        self._Q[action] = (1-alpha) * old_Q + alpha * reward
+        self._N[action] = self._N[action] + 1
 
 class UcbBanditLearner(Learner):
     """This is an implementation of Auer et al. (2002) UCB1-Tuned algorithm.
@@ -143,80 +84,62 @@ class UcbBanditLearner(Learner):
         """Instantiate a UcbBanditLearner."""
 
         #these variable names were selected for easier comparison with the original paper 
-        self._init_a: int = 0
         self._t     : int = 0
-        self._s     : Dict[Action, int           ] = defaultdict(int)
         self._m     : Dict[Action, float         ] = {}
-        self._v     : Dict[Action, OnlineVariance] = defaultdict(OnlineVariance)
-
-    @property
-    def family(self) -> str:
-        """The family of the learner.
-
-        See the base class for more information
-        """
-        return "bandit_UCB"
+        self._s     : Dict[Action, int           ] = {}
+        self._v     : Dict[Action, OnlineVariance] = {}
 
     @property
     def params(self) -> Dict[str, Any]:
         """The parameters of the learner.
-        
+
         See the base class for more information
         """
-        return { }
+        return { "family": "bandit_UCB" }
 
-    def predict(self, key: Key, context: Context, actions: Sequence[Action]) -> Sequence[float]:
+    def predict(self, context: Context, actions: Sequence[Action]) -> Probs:
         """Determine a PMF with which to select the given actions.
 
         Args:
-            key: The key identifying the interaction we are choosing for.
             context: The context we're currently in. See the base class for more information.
             actions: The actions to choose from. See the base class for more information.
 
         Returns:
             The probability of taking each action. See the base class for more information.
         """
+        self._t += 1
+        never_observed_actions = [ a for a in actions if a not in self._m ]
 
-        actions = [ self._key(a) for a in actions ]
-
-        #initialize by playing every action once
-        if self._init_a < len(actions):
-            self._init_a += 1
-            return [ int(i == (self._init_a-1)) for i in range(len(actions)) ]
-
+        if never_observed_actions:
+            max_actions = never_observed_actions
         else:
-            values      = [ self._m[a] + self._Avg_R_UCB(a) if a in self._m else None for a in actions ]
-            max_value   = None if set(values) == {None} else max(v for v in values if v is not None)
-            max_indexes = [i for i in range(len(values)) if values[i]==max_value]
+            values      = [ self._m[a] + self._Avg_R_UCB(a) for a in actions ]
+            max_value   = max(values)
+            max_actions = [ a for a,v in zip(actions,values) if v==max_value ]
 
-            return [ int(i in max_indexes)/len(max_indexes) for i in range(len(actions)) ]
+        return [ int(action in max_actions)/len(max_actions) for action in actions ]
 
-    def learn(self, key: Key, context: Context, action: Action, reward: float, probability: float) -> None:
+    def learn(self, context: Context, action: Action, reward: float, probability: float, info: Info) -> None:
         """Learn from the given interaction.
 
         Args:
-            key: The key identifying the interaction this observed reward came from.
             context: The context we're learning about. See the base class for more information.
             action: The action that was selected in the context. See the base class for more information.
             reward: The reward that was gained from the action. See the base class for more information.
-            probability: The probability that the given action was taken.
+            probability: The probability with wich the given action was selected.
+            info: Optional information provided during prediction step for use in learning.
         """
 
         assert 0 <= reward and reward <= 1, "This algorithm assumes that reward has support in [0,1]."
 
-        action = self._key(action)
-
         if action not in self._m:
             self._m[action] = reward
+            self._s[action] = 1
+            self._v[action] = OnlineVariance()
         else:
             self._m[action] = (1-1/self._s[action]) * self._m[action] + 1/self._s[action] * reward
-
-        self._t         += 1
-        self._s[action] += 1
-        self._v[action].update(reward)
-
-    def _key(self, action: Action) -> Hashable:
-        return tuple(action.items()) if isinstance(action,dict) else action
+            self._s[action] += 1
+            self._v[action].update(reward)
 
     def _Avg_R_UCB(self, action: Action) -> float:
         """Produce the estimated upper confidence bound (UCB) for E[R|A].
@@ -248,4 +171,4 @@ class UcbBanditLearner(Learner):
         """
         ln = math.log; t = self._t; s = self._s[action]; var = self._v[action].variance
 
-        return var + math.sqrt(2*ln(t)/s)
\ No newline at end of file
+        return var + math.sqrt(2*ln(t)/s)
