@@ -86,6 +86,11 @@ class CobaRandom:
         return l
 
     def random(self) -> float:
+        """Generate a uniform random number in [0,1].
+
+        Returns:
+            The generated random number in [0,1].
+        """
         return self.randoms(1)[0]
 
     def randint(self, a:int, b:int) -> int:
@@ -130,8 +135,8 @@ class CobaRandom:
             The `n` generated random numbers in [0,m-1].
         """
         
-        if n <= 0 or not isinstance(n, int):
-            raise ValueError("n must be an integer greater than 0")
+        if n < 0 or not isinstance(n, int):
+            raise ValueError("n must be an integer greater than or equal 0")
 
         numbers: List[int] = []
 
