@@ -3,10 +3,19 @@
 import sys
 import os
 
+from collections import defaultdict
 from io import UnsupportedOperation
 from contextlib import contextmanager
 from typing import IO
 
+from coba.exceptions import CobaException
+
+def coba_exit():
+    try:
+        exit() #this is my attempt at handling CPython (where this works) vs IPython (e.g., Jupyter Notebooks)
+    except NameError:
+        raise CobaException() #CobaException has been specially made so that it outputs no TB in Jupyter Notebook
+
 @contextmanager
 def redirect_stderr(to: IO[str]):
     """Redirect stdout for both C and Python.
@@ -63,7 +72,7 @@ def redirect_stderr(to: IO[str]):
 class PackageChecker:
 
     @staticmethod
-    def matplotlib(caller_name: str) -> None:
+    def matplotlib(caller_name: str, silent: bool = False) -> None:
         """Raise ImportError with detailed error message if matplotlib is not installed.
 
         Functionality requiring matplotlib should call this helper and then lazily import.
@@ -77,34 +86,29 @@ class PackageChecker:
         """
         try:
             import matplotlib # type: ignore
-        except ImportError as e:
-            raise ImportError(
-                caller_name + " requires matplotlib. You can "
-                "install matplotlib with `pip install matplotlib`."
-            ) from e
+        except ImportError:
+            PackageChecker._handle_import_error(caller_name, "matplotlib", silent)
 
     @staticmethod
-    def vowpalwabbit(caller_name: str) -> None:
+    def vowpalwabbit(caller_name: str, silent: bool = False) -> None:
         """Raise ImportError with detailed error message if vowpalwabbit is not installed.
 
         Functionality requiring vowpalwabbit should call this helper and then lazily import.
 
         Args:    
-            caller_name: The name of the caller that requires matplotlib.
+            caller_name: The name of the caller that requires vowpalwabbit.
 
         Remarks:
             This pattern was inspired by sklearn (see `PackageChecker.matplotlib` for more information).
         """
-        try:
+
+        try:            
             import vowpalwabbit # type: ignore
         except ImportError as e:
-            raise ImportError(
-                caller_name + " requires vowpalwabbit. You can "
-                "install vowpalwabbit with `pip install vowpalwabbit`."
-            ) from e
+            PackageChecker._handle_import_error(caller_name, "vowpalwabbit", silent)
 
     @staticmethod
-    def pandas(caller_name: str) -> None:
+    def pandas(caller_name: str, silent: bool = False) -> None:
         """Raise ImportError with detailed error message if pandas is not installed.
 
         Functionality requiring pandas should call this helper and then lazily import.
@@ -117,14 +121,11 @@ class PackageChecker:
         """
         try:
             import pandas # type: ignore
-        except ImportError as e:
-            raise ImportError(
-                caller_name + " requires pandas. You can "
-                "install pandas with `pip install pandas`."
-            ) from e
+        except ImportError:
+            PackageChecker._handle_import_error(caller_name, "pandas", silent)
 
     @staticmethod
-    def numpy(caller_name: str) -> None:
+    def numpy(caller_name: str, silent: bool = False) -> None:
         """Raise ImportError with detailed error message if numpy is not installed.
 
         Functionality requiring numpy should call this helper and then lazily import.
@@ -137,28 +138,48 @@ class PackageChecker:
         """
         try:
             import numpy # type: ignore
-        except ImportError as e:
-            raise ImportError(
-                caller_name + " requires numpy. You can "
-                "install numpy with `pip install numpy`."
-            ) from e
+        except ImportError:
+            PackageChecker._handle_import_error(caller_name, "numpy", silent)
 
     @staticmethod
-    def sklearn(caller_name: str) -> None:
-        """Raise ImportError with detailed error message if numpy is not installed.
+    def sklearn(caller_name: str, silent: bool = False) -> None:
+        """Raise ImportError with detailed error message if sklearn is not installed.
 
-        Functionality requiring numpy should call this helper and then lazily import.
+        Functionality requiring sklearn should call this helper and then lazily import.
 
         Args:
-            caller_name: The name of the caller that requires numpy.
+            caller_name: The name of the caller that requires sklearn.
 
         Remarks:
             This pattern was inspired by sklearn (see `PackageChecker.matplotlib` for more information).
         """
         try:
             import sklearn # type: ignore
-        except ImportError as e:
-            raise ImportError(
-                caller_name + " requires sklearn. You can "
-                "install sklearn with `pip install sklearn`."
-            ) from e
\ No newline at end of file
+        except ImportError:
+            PackageChecker._handle_import_error(caller_name, "scikit-learn", silent)
+
+    def _handle_import_error(caller_name:str, pkg_name:str, silent:bool):
+        if not silent:
+            print(f"ERROR: {caller_name} requires the {pkg_name} package. You can install this package via `pip install {pkg_name}`.")
+        coba_exit()
+
+class HashableDict(dict):
+    def __init__(self, *args, **kwargs) -> None:
+        super().__init__(*args, **kwargs)
+
+        self._hash = hash(tuple(self.items()))
+
+    def __hash__(self) -> int:
+
+        assert self._hash == hash(tuple(self.items()))
+
+        return self._hash
+
+class KeyDefaultDict(defaultdict):
+    def __missing__(self, key):
+        if self.default_factory is None:
+            raise KeyError( key )
+        else:
+            value = self.default_factory(key)
+            self[key] = value
+            return value
\ No newline at end of file
